/* Based on http://www.kernel.org/pub/linux/libs/pam/pre/doc/rfc86.0.txt.gz 
 *
 * This hereby is dedicated into the Public Domain by
 * Valentin `Tino' Hilbig, webmaster@20k.de
 *
 * PAM needs following bible sized routine just to verify a password.
 * Properly planned this should go into 1 line .. OR LESS!
 *
 * Don't forget to add into /etc/pam.d/ the file called after the first parameter.
 * It looks like this:
 * #%PAM-1.0
 * auth       required	pam_unix2.so
 * account    required	pam_unix2.so
 *
 * 2004-01-10 Tino: Version 1
 */
#include <stdio.h>

#ifndef USE_PAM

/* Untested code
*/

#include <string.h>
#include <pwd.h>
#include <crypt.h>

int
password_verify(const char *app, const char *user, const char *pass)
{
  struct passwd *pw;

  /* authenticate password */
  pw = getpwnam(user);
  return !pw || strcmp(crypt(pass, pw->pw_passwd), pw->pw_passwd);
}

#else

#include <stdlib.h>
#include <string.h>
#include <security/pam_appl.h>

#if 0
#define DP(X)	do { } while (0)
#else
#define DP(X)	do { dprintf X; } while (0)
#include <stdarg.h>
static void
dprintf(const char *f, ...)
{
  static FILE	*fd;
  va_list	list;

  if (!fd)
    fd	= fopen("/tmp/password_verify.debug", "a");
  if (!fd)
    return;
  va_start(list, f);
  vfprintf(fd, f, list);
  va_end(list);
}
#endif

struct password_verify_data
  {
    const char	*user, *pass;
  };

/* In this routine the RFC is buggy as hell.
 * Had to look in mod_auth_pam.c to find something which is working.
 * This leads to the question,
 * how complex can anything be made until it becomes absurd?
 */
static int
password_verify_conv(int num_msg, const struct pam_message **msg,
		     struct pam_response **response, void *appdata_ptr)
{
  struct pam_response		*r	= 0;
  struct password_verify_data	*u	= appdata_ptr;
  int				i;
  
  DP(("password_verify_conv(%d,%p,%p,%p)\n", num_msg, msg, response, appdata_ptr));
  if (num_msg<0 || !msg || !response || !u)
    return PAM_CONV_ERR;

  if ((r=malloc(num_msg * sizeof *r))==NULL)
    return PAM_CONV_ERR;

  DP(("password_verify_conv user=%s pw=%s\n", u->user, u->pass));
  for (i=0; i<num_msg; i++)
    {
      const char	*p;

      DP(("password_verify_conv msg=%d\n", msg[i]->msg_style));
      switch (msg[i] ? msg[i]->msg_style : -1)
	{
	case -1:
	default:
	  free(r);
	  return PAM_CONV_ERR;

	case PAM_PROMPT_ECHO_ON:
	  p	= u->user;
	  break;

	case PAM_PROMPT_ECHO_OFF:
	  p	= u->pass;
	  break;
	}
      DP(("password_verify_conv ans=%s\n", p));
      r[i].resp		= strdup(p);
      r[i].resp_retcode	= PAM_SUCCESS;
    }
  
  DP(("password_verify_conv SUCCESS\n"));
  *response	= r;
  return PAM_SUCCESS;
}

int
password_verify(const char *app, const char *user, const char *pw)
{
  struct password_verify_data	u	= { user, pw };
  struct pam_conv		pam_conv= { password_verify_conv, &u};
  pam_handle_t			*pamh;
  int				ret;

  ret	= 0;
  pamh	= 0;

  /* Open PAM
   */
  if (pam_start(app, user, &pam_conv, &pamh)!=PAM_SUCCESS)
    ret	= -1;

  /* Check if I am authenticated.
   */
  if (!ret && pam_authenticate(pamh, PAM_DISALLOW_NULL_AUTHTOK)!=PAM_SUCCESS)
    ret	= 1;

  /* Check if the authentication is valid. (Why is this needed???)
   */
  if (!ret && pam_acct_mgmt(pamh, PAM_DISALLOW_NULL_AUTHTOK)!=PAM_SUCCESS)
    ret	= 2;

#if 0
  /* We cannot open a session, as we do not close the session later
   */
  if (!ret && pam_open_session(pamh, 0)!=PAM_SUCCESS)
    ret	= 3;
#endif

  if (pamh)
    pam_end(pamh, PAM_SUCCESS);
  return ret;
}

#endif
